--[[
This plug in calculates dps values based on the calculator written by
Toskk (http://druid.wikispaces.com/ToskksDPSGearMethod),
Tangedyn (http://tangedyn.9tribes.net/cat-dps/catdps-calc/)
using research gathered by many others.
--]]



local GCD_TIME = 1.0
local EPS=10
local SPE=.1
local GCD_ENERGY = GCD_TIME*EPS
local GLANCING_MODIFIER = .7
local GLANCING_CHANCE = .25
local MALORNE_ENERGY =20
local MALORNE_PROC_CHANCE = .04
local MALORNE_ENERGY_PER_HIT =MALORNE_PROC_CHANCE*MALORNE_ENERGY
local OOC_PROC = 2/60
local RIP_TIME = 12
local RIP_COST = 30


local MALORNE_HARNESS = "MalorneHarness"
local WASTEWALKER_ARMOR = "WastewalkerArmor"
local PRIMAL_INTENT = "PrimalIntent"
local STRENGTH_OF_THE_CLEFTHOOF = "StrengthoftheClefthoof"


local function print(x)
	if(type(x)=="table") then
		for k,v in pairs(x) do
			ChatFrame1:AddMessage(k.." = "..tostring(v))
		end
	else
		ChatFrame1:AddMessage(x)
	end
end

local Options = {
	Enemy ={
		Mitigation	= CatPoints:MitigationFromArmor(7700),--.15,
		Miss				= .09,
		Dodge				= .05,
	},
	ShredsPerMangle=3,
	BaseStat='ATTACKPOWER',

	derivedStats = {
		[MALORNE_HARNESS] = function()
			local eq = CatPoints.EquippedSets[MALORNE_HARNESS] or 0
			return eq
		end,
		[WASTEWALKER_ARMOR] = function()
			local eq = CatPoints.EquippedSets[WASTEWALKER_ARMOR] or 0
			return eq
		end,
		[PRIMAL_INTENT] = function()
			local eq = CatPoints.EquippedSets[PRIMAL_INTENT] or 0
			return eq
		end,
		Idol = function()
			return CatPoints:GetIdol()
		end,
		Haste = function()
			return (GetCombatRating(CR_HASTE_MELEE))
		end,
		Crit = function()
			return GetCritChance()/100
		end,
		Hit= function()
			return (GetCombatRating(CR_HIT_MELEE))
		end,
		Expertise = function()
			return (GetCombatRating(CR_WEAPON_SKILL))
		end,
		WeaponDamage = function()
			local _, _, _, _, dmg, _, _ = UnitDamage("player")
			return dmg
		end,
		AttackPower = function()
			local base, posBuff, negBuff = UnitAttackPower("player")
			return base+posBuff+negBuff
		end,
		AttackSpeed = function(player)
			return 1/(1+player.Haste/15.76/100)
		end,
		DodgeReduction= function(player)
			return player.Expertise*52/82/2.5*.25/100
		end,
		Dodge = function(player)
			local dodge = player.Enemy.Dodge - player.DodgeReduction
			if(dodge < 0) then
				dodge = 0
			end
			return dodge
		end,
		Melee=function(player)
			return player.WeaponDamage + (768+player.AttackPower)/14
		end,
		MeleeCrit=function(player)
			return player.Melee*player.CritMultiplier
		end,
		MissReduction=function(player)
			return player.Hit*52/82/10/100
		end,
		Miss = function(player)
			local miss = player.Enemy.Miss-player.MissReduction
			if(miss < 0 ) then
				miss = 0
			end
			return miss
		end,
		CritMultiplier = function(player)
			local mult = (2 * (1 + (player.PredatoryInstincts * 0.02)))
			if(player.Diamond) then
				mult = mult *1.03
			end
			return mult
		end,
		MangleDamage=function(player)
			local sf = (1 + player.SavageFury * 0.1)
			local mangle= (player.Melee * 1.6 + 264)
			if(player.Idol == "IdoloftheWild") then
				mangle = mangle + 24
			end

			return mangle*sf
		end,
		MangleCritDamage=function(player)
			local mangle = player.MangleDamage

			return mangle*player.CritMultiplier
		end,
		MangleCost = function(player)
			return 45-player.Ferocity - ((player.Thunderheart>=2 and 5) or 0)
		end,
		ShredDamage=function(player)
			local shred= (player.Melee * 2.25 + 405)
			if(player.Idol == "EverbloomIdol") then
				shred = shred + 88
			end
			if(player.Nordrassil >=4) then
				shred = shred + 75
			end

			return shred
		end,
		ShredCritDamage=function(player)
			local shred = player.ShredDamage

			return shred*player.CritMultiplier
		end,
		ShredCost = function(player)
			return 60-player.ShreddingAttacks*9
		end,
		ShredMissCost = function(player)
			return player.ShredCost *.2
		end,
		MangleMissCost = function(player)
			return player.MangleCost *.2
		end,
		RipMissCost = function(player)
			return player.RipCost
		end,
		MangleAmortizedCost = function(player)
			local miss = player.Dodge+player.Miss
			return 1/(1-miss)*(player.MangleCost*(1-miss)+player.MangleMissCost*miss)
		end,
		ShredAmortizedCost = function(player)
			local miss = player.Dodge+player.Miss

			-- QFT why is this different from mangle? 1/(1-miss)*
			return (player.ShredCost*(1-miss)+player.ShredMissCost*miss)
		end,
		RipAmortizedCost = function(player)
			local miss = player.Dodge+player.Miss

			return 1/(1-miss)*RIP_COST
		end,
	},
	derivedFunctions ={
		RipDamage = function(player, cp)
			local _apmult = .06*cp
			if(_apmult > .24) then
				_apmult = .24
			end

			damage = (148+281*cp + _apmult*player.AttackPower+(player.Idol=="IdolofFeralShadows" and 4*cp*6 or 0)) * (player.Thunderheart>=4 and 1.15 or 1)
			return damage
		end
	}
}



function Options:display()

	return true
end



function Options:applyBonus(bonuses, considerSet)

	local relevant = nil
	local s = self:new({})
	if(#bonuses == 0) then
		relevant = {}
	end

	setmetatable(bonuses, {__index = function(_, name) if(relevant) then relevant[name]=true end return 0 end})

	local SetBonuses = {}
	setmetatable(SetBonuses, {__index = function(_, name) return 0 end})

	if(considerSet) then
		CatPoints:ConsiderSet(s, MALORNE_HARNESS, bonuses, SetBonuses, 4, 'STR', 30)
		CatPoints:ConsiderSet(s, WASTEWALKER_ARMOR, bonuses, SetBonuses, 2, 'CR_HIT', 35)
		CatPoints:ConsiderSet(s, PRIMAL_INTENT, bonuses, SetBonuses, 3, 'ATTACKPOWER', 40)
		CatPoints:ConsiderSet(s, STRENGTH_OF_THE_CLEFTHOOF, bonuses, SetBonuses, 3, 'STR', 20)
	end

	--AP
	s.AttackPower = s.AttackPower + (bonuses.ATTACKPOWER + bonuses.ATTACKPOWERFERAL+SetBonuses.ATTACKPOWER) *( 1 + s.HeartoftheWild*.02 )



	--STR
	s.AttackPower = s.AttackPower + ((bonuses.STR+SetBonuses.STR )*(1+s.SurvivaloftheFittest*.01)*2) *( 1 + s.HeartoftheWild*.02 )*(1+ .1*s.Kings)

	--AGI->AP
	--print(s.AttackPower)
	s.AttackPower = s.AttackPower + (bonuses.AGI*(1+s.SurvivaloftheFittest*.01)) *( 1 + s.HeartoftheWild*.02 )*(1+ .1*s.Kings)

	--print(s.AttackPower)
	--AGI->Crit
	s.Crit = s.Crit + (bonuses.AGI*(1+s.SurvivaloftheFittest*.01))/2500*(1+ .1*s.Kings)

	--print(s.Crit)
	--CR_CRIT->Crit
	s.Crit = s.Crit + (bonuses.CR_CRIT)*(52/82/14)/100

	--CatPoints:Print(s.AttackPower, s.Crit)
	s.Penetration = s.Penetration + bonuses.IGNOREARMOR
	s.Hit = s.Hit + bonuses.CR_HIT + SetBonuses.CR_HIT
	s.Haste = s.Haste + bonuses.CR_HASTE
	s.WeaponDamage = s.WeaponDamage + bonuses.MELEEDMG
	s.Expertise = s.Expertise + bonuses.CR_EXPERTISE +bonuses.CR_WEAPON_FERAL

	if(considerSet) then
		s.Nordrassil = s.Nordrassil + bonuses["Nordrassil Harness"]
		s.Thunderheart = s.Thunderheart + bonuses["Thunderheart Harness"]
	end

	if(bonuses.Idol and bonuses.Idol ~= 0) then
		s.Idol = bonuses.Idol
	end

	return s,relevant
end



function Options:calculate()
	local enemy = self.Enemy
	local baseArmor = CatPoints:ArmorFromMitigation(enemy.Mitigation, self.Penetration)
	local actualMitigation = CatPoints:MitigationFromArmor(baseArmor)

	local actualMiss = self.Miss

	local actualDodge = self.Dodge

	local totalMiss = (actualMiss + actualDodge)
	local actualHit = 1- totalMiss

	if(TraceStep) then
		TraceStep.totalMiss = self.totalMiss
	end

	local actualCrit = self.Crit
	if(actualCrit>.75) then
		actualCrit = .75
	end

	local whiteCrit = actualCrit
	if(whiteCrit > (.75-totalMiss)) then
		whiteCrit = .75-totalMiss
	end

	local cycleTime = GCD_TIME
	local totalEnergy = self.MangleCost + GCD_ENERGY
	local segmentTime = GCD_TIME

	if(TraceStep) then
		TraceStep.Melee = self.Melee
		TraceStep.totalMiss = totalMiss
		TraceStep.whiteCrit = whiteCrit
		TraceStep.PredatoryInstincts = self.PredatoryInstincts
		TraceStep.MangleDamage = self.MangleDamage
	end

	local meleeAvgAttack = 1/self.AttackSpeed * self.Melee *
		(GLANCING_MODIFIER * GLANCING_CHANCE + ((1-GLANCING_CHANCE)- totalMiss - whiteCrit) + self.CritMultiplier * (whiteCrit))

	local meleeDmg = segmentTime * meleeAvgAttack
	if(TraceStep) then	TraceStep.meleeDmg1 = meleeDmg end

	if(self[MALORNE_HARNESS] >= 2) then
		totalEnergy = totalEnergy + segmentTime/self.AttackSpeed * MALORNE_ENERGY_PER_HIT
	end

	totalEnergy = totalEnergy-self.MangleAmortizedCost*(1-segmentTime*OOC_PROC)
	if(TraceStep) then	TraceStep.totalEnergyAfterMangle = totalEnergy end

	local totalMangleDamage = self.MangleDamage*(1-actualCrit) + self.MangleCritDamage*actualCrit
	if(TraceStep) then	TraceStep.totalMangleDamage = totalMangleDamage end

	local attacks = (4-3*actualCrit+2*actualCrit*actualCrit-actualCrit*actualCrit*actualCrit)/actualHit
	if(TraceStep) then	TraceStep.attacks = attacks end

	local nShreds = attacks-1/actualHit

	if(self[MALORNE_HARNESS] >=2) then
		totalEnergy = totalEnergy + nShreds*MALORNE_ENERGY_PER_HIT
	end

	segmentTime = (nShreds*self.ShredAmortizedCost*(1-nShreds*self.ShredAmortizedCost/10/nShreds*OOC_PROC) - totalEnergy)/10

	if(self[MALORNE_HARNESS] >=2) then
		segmentTime = segmentTime-segmentTime/self.AttackSpeed*MALORNE_ENERGY_PER_HIT*SPE
	end
	cycleTime = cycleTime + segmentTime
	if(TraceStep) then
		TraceStep.shredsSegment = segmentTime
		TraceStep.ShredAmortizedCost = self.ShredAmortizedCost
	end

	meleeDmg = meleeDmg + segmentTime * meleeAvgAttack
	if(TraceStep) then	TraceStep.meleeDmgAfterShreds = meleeDmg end

	local totalShredDamage = nShreds*self.ShredDamage*1.3*((1-totalMiss)*((1-actualCrit) +self.CritMultiplier*actualCrit))
	if(segmentTime > RIP_TIME) then
		--QFT non mangled shreds?
		totalShredDamage = ((RIP_TIME / segmentTime) * totalShredDamage) + ((1 - RIP_TIME / segmentTime) * totalShredDamage/1.3)
	end
	if(TraceStep) then	TraceStep.totalShredDamage = totalShredDamage end


	totalEnergy = 0
	if(self[MALORNE_HARNESS] >= 2) then
		totalEnergy = totalEnergy + MALORNE_ENERGY_PER_HIT
	end
	-- Wait for enough energy for rip + mangle
	segmentTime = (self.MangleCost+self.RipAmortizedCost * (1- self.RipAmortizedCost*SPE*OOC_PROC) - totalEnergy)*SPE
	if(self[MALORNE_HARNESS] >= 2) then
		segmentTime = segmentTime-segmentTime/self.AttackSpeed*MALORNE_ENERGY_PER_HIT*SPE
	end
	cycleTime = cycleTime+segmentTime
	if(TraceStep) then
		TraceStep.waitTime = segmentTime
		TraceStep.cycleTime = cycleTime
	end

	meleeDmg = meleeDmg + segmentTime * meleeAvgAttack
	if(TraceStep) then	TraceStep.meleeDmgBeforeRip = meleeDmg end

	local rip5Total = actualCrit - actualCrit*actualCrit+ actualCrit*actualCrit*actualCrit - actualCrit*actualCrit*actualCrit*actualCrit
	local rip4Total = 1-rip5Total

	if(cycleTime < RIP_TIME) then
		--QFT study this
		segmentTime = RIP_TIME - cycleTime
		cycleTime = cycleTime + segmentTime
		totalEnergy = segmentTime*EPS
		meleeDmg = meleeDmg + segmentTime * meleeAvgAttack
		if(self[MALORNE_HARNESS] >= 2) then
			segmentTime = segmentTime/self.AttackSpeed*MALORNE_ENERGY_PER_HIT
		end
		local extraShreds = totalEnergy / (self.ShredAmortizedCost * (1-segmentTime*OOC_PROC))
		totalShredDamage = totalShredDamage+extraShreds*self.ShredDamage*1.3*((1-totalMiss)*((1-actualCrit) +self.CritMultiplier*actualCrit))

		rip5Total = rip5Total+extraShreds*rip4Total*rip4Total
		rip4Total = rip4Total-extraShreds*rip4Total*rip4Total
	end

	local ripDamage = 1.3*(rip5Total*self.RipDamage(5)+rip4Total*self.RipDamage(4))
	if(TraceStep) then
		TraceStep.ripDamage = ripDamage
	end

	local dps = ((totalMangleDamage+totalShredDamage+meleeDmg)*actualMitigation+ripDamage)/cycleTime
	if(TraceStep) then
		TraceStep.finaltotalMangleDamage = totalMangleDamage
		TraceStep.finaltotalShredDamage = totalShredDamage
		TraceStep.finalmeleeDmg = meleeDmg
		TraceStep.actualMitigation = actualMitigation
	end
	return dps* (1+.02*self.Naturalist)
end




local Player = CatPoints:CreateTable("CatDPS", Options)



