--[[
This addon is a harness for adding point values to item tooltips.
The definition of the calculations are defined in external plug
in files.

Enchants on items being compared are removed for the calculation (it is
assumed you would put the same enchant on both items.  Gems are taken
into account, but empty slots are given no value.  I have added an addon
to files.wowace.com that will create item links with gems of your choice
filled in, which is handy for comparing using this addon.

To get set bonus detection to work, I had to modify my local copy of
ItemBonusLib.

Known Bugs:
- This does put some extra values in unwanted tooltips (like junk items).
- Comparisons are always done vs the first slot of items that can go in finger 2, trinket 2 and offhand.
- Set bonuses require a modification of ItemBonusLib to be detected.
- Idol detection needs some work.
- There is a bog that happens some times where the values some how come out

Future work:
- Add more Idols
- Add more set bonuses
- Restrict bonus display to form
- Integrate with Bedazzler
- Add values for on use trinkets and procs
]]




local EOL = ''


function prettyPrint(t, indent, call)
	if(not indent) then
		indent=""
	end
	if(type(t)=='table') then
		local str ="{"..EOL
		for k,v in pairs(t) do
			if(type(v)~='function' and k ~='parent' and k ~= 'relevant') then
				if(k=='derivedStats') then
					str = str.."{"..EOL
					for m, n in pairs(v) do
						str=str..indent.."   "..(m or "nil").." = "..(m and t[m] or "nil")..","..EOL
					end
					str=str..indent.." }"
				else
					str=str..indent.." "..k.." = "..prettyPrint(v, indent.." ")..","..EOL
				end
			end
		end
		return str..indent.."}"
	else
		return tostring(t)
	end
end



local Player={

	--Enter these manually
	Kings								= 0,
	Penetration					= 0,
	Diamond							= false,

	--currently these are used, but ItemBonusLib has no API for
	--returning these values.  They should be set manually for
	--now until a patch can be made
	Thunderheart 				= 0,
	Nordrassil					= 0,
	MalorneHarness			= 0,
	WastewalkerArmor		= 0,
	PrimalIntent				= 0,
	StrengthoftheClefthoof = 0,

	--These should be detected
	SavageFury					= 0,--2,
	Ferocity						= 0,--5,
	ShreddingAttacks		= 0,--2,
	PredatoryInstincts	= 0,--5,
	Naturalist					= 0,--5,
	HeartoftheWild			= 0,--5,
	SurvivaloftheFittest= 0,--3,
	OmenofClarity				= 0,--1,
	ThickHide						= 0,--3

	Idol 								= 'None',
}


local function print(x, ...)
	if(type(x)=="table") then
		for k,v in pairs(x) do
			ChatFrame1:AddMessage(k.." = "..tostring(v))
		end
	else
		ChatFrame1:AddMessage(x)
	end
	if(#{...} > 0 ) then
		print(...)
	end
end


CatPoints = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceConsole-2.0", "AceHook-2.1")

CatPoints.Bonus = AceLibrary("ItemBonusLib-1.0")


-- Data
INVTYPE_WEAPON_OTHER = INVTYPE_WEAPON.."_other";
INVTYPE_FINGER_OTHER = INVTYPE_FINGER.."_other";
INVTYPE_TRINKET_OTHER = INVTYPE_TRINKET.."_other";

local invType = {
INVTYPE_HEAD=1,
INVTYPE_NECK=2,
INVTYPE_SHOULDER=3,
INVTYPE_BODY=4,
INVTYPE_CHEST=5,
INVTYPE_WAIST=6,
INVTYPE_LEGS=7,
INVTYPE_FEET=8,
INVTYPE_WRIST=9,
INVTYPE_HAND=10,
INVTYPE_FINGER=11,
INVTYPE_TRINKET=13,
INVTYPE_CLOAK=15,
INVTYPE_WEAPON=16,
INVTYPE_2HWEAPON=16,
INVTYPE_WEAPONMAINHAND=16,
INVTYPE_HOLDABLE=17,
INVTYPE_RELIC =18
}

local itemCache =
{
	link = nil,
	eqlink =nil,
	lines = {},
}

local function ColorValue(value)
	local total
	if(type(value) == "number") then
		if(value > 0) then
			total = string.format("|cff00ff00%.2f|r", value)
		elseif (value < 0) then
			total = string.format("|cffff0000%.2f|r", value)
		end
	else
		total = tostring(value)
	end
	return total
end


function CatPoints:AddPointsLine(frame, caption, bonuses, setBonuses, add)
	local total

	if(setBonuses and setBonuses ~= bonuses) then
		total = string.format("(%s) %s", ColorValue(bonuses), ColorValue(setBonuses))
	else
		total = ColorValue(bonuses)
	end

	if(total) then
		frame:AddDoubleLine(caption, total,1,1,1,1,1,1)
		if(add) then
			tinsert(itemCache.lines, {caption, total})
		end
	end
end

local function BareID(itemString)
	if(itemString) then

		local prefix,itemId,enchantId,g1,g2,g3,g4,suffixId,uniqueId = strsplit(":", itemString)
		if(prefix and itemId and 0 and g1 and g2 and g3 and g4 and suffixId and uniqueId) then
			return strjoin(":", prefix, itemId, 0, g1,g2,g3,g4, suffixId, uniqueId)
		end
	end
	return itemString
end

local SpecialItemMap =
{
	[33831] = {Name="Berserker's Call", Bonuses = {ATTACKPOWER=360*20/120}},
	[31856] = {Name="Darkmoon Card: Crusade", Bonuses = {ATTACKPOWER=120}},	-- <120 for short fights or in and out fights
	[29383] = {Name="Bloodlust Brooch", Bonuses={ATTACKPOWER=278*20/120}},
	[28830] = {Name="Dragonspine Trophy", Bonuses={CR_HASTE=325*.25}},			-- 25% uptime
	[30627] = {Name="Tsunami Talisman", Bonuses={ATTACKPOWER=340*10/60}}, 	-- 1PPM
	[32654] = {Name="Crystalforged Trinket", Bonuses={ATTACKPOWER=216*10/60}},
	[23041] = {Name="Slayer's Crest", Bonuses={ATTACKPOWER=260*20/120}},
	[32505] = {Name="Madness of the Betrayer", Bonuses={IGNOREARMOR=300/3}},
	[28034] = {Name="Hourglass of the Unraveller", Bonuses={ATTACKPOWER=300*10/60}},
	[28288] = {Name="Abacus of Violent Odds", Bonuses={CR_HASTE=260*10/120}},
	[29776] = {Name="Core of Ar'kelos", Bonuses={ATTACKPOWER=200*20/120}},
	[32257] = {Name="Idol of the White Stag", Bonuses ={ATTACKPOWER=94}},

	[32658] = {Name="Badge of Tenacity", Bonuses ={AGI=150*20/120}},

	[27770] = {Name="Argussian Compass", Bonuses={HEALTH=1150*20/120}},
	[7838] = {Name="Arena Grandmaster", Bonuses={HEALTH=1000*20/120}},
	[32501] = {Name="Shadowmoon Insignia", Bonuses={HEALTH=1750*20/180}}, --ug i dont like this one
	[30620] = {Name="Spyglass of the Hidden Fleet", Bonuses={HEALTH=1300*12/120}},
	[24376] = {Name="Runed Fungalcap", Bonuses={HEALTH=440*20/120}},
	[28528] = {Name="Moroes' Lucky Pocket Watch", Bonuses={CR_DODGE=300*10/120}},
	[30629] = {Name="Scarab of Displacement", Bonuses={CR_DEFENSE=165*15/180}},
	[27891] = {Name="Adamantine Figurine", Bonuses={ARMOR_BONUS=1280*20/120}},


}

function CatPoints:MergeSpecial(bonuses, link)
	if(bonuses and link) then
		local _,itemId=strsplit(":", link)
		itemId=tonumber(itemId)
		if(itemId and SpecialItemMap[itemId]) then
			local b = SpecialItemMap[itemId].Bonuses
			for k,v in pairs(b) do
				bonuses[k]=(bonuses[k] or 0) + v
			end
		end
	end
end

function CatPoints:GetIdol(link)
	local idol = link or GetInventoryItemLink("player",GetInventorySlotInfo("RangedSlot"))
	if(idol and type(idol) == 'string') then
		idol = GetItemInfo(idol):gsub("%W", "")
	end

	return idol or 'None'
end

function CatPoints:ConsiderSet(s, name, bonuses, SetBonuses, value, stat, amount)
	if(s[name] < value and s[name] + bonuses[name] >= value) then
		SetBonuses[stat] = amount
	elseif(s[name] >= value and s[name] + bonuses[name] < value) then
		SetBonuses[stat] = -amount
	end
	s[name] = s[name] + bonuses[name]
end


function Player:display()
	return true
end

function Player:compare(netBonus)
	local base = self:calculate()
	local withSet = self:applyBonus(netBonus, true)
	local noSet = self:applyBonus(netBonus, false)
	return withSet:calculate()-base , noSet:calculate()-base
end

function Player:equivalence(noDisplay)

	local player, relevant = self:applyBonus({}, true)
	local base = player:calculate()

	local calculated = {}


	for stat in pairs(relevant) do
		local p = player:applyBonus({[stat]=1}, true)
		local value = p:calculate()-base
		if(value ~= 0) then
			calculated[stat]= value
		end
	end

	local relative = calculated[player.BaseStat]

	for stat,value in pairs(calculated) do
		calculated[stat]=value/relative
	end

	calculated.Points = base


	if(not noDisplay) then
		for k,v in pairs(calculated) do
			print(string.format("% 12.5f = %s", v,k))
		end
	end
	return calculated
end

function Player:new(stats, tableName)

	if(not stats) then
		stats = {}
	end
	stats.parent=self


	setmetatable(stats, {__index=function(t, name)
		--print("looking up "..name.." in "..tostring(t))
		if(rawget(t, "derivedStats") and t.derivedStats[name]) then
			--print("derivedStat")
			local value = t.derivedStats[name](stats)
			--t[name]= value
			return value
		elseif(rawget(t, "derivedFunctions") and t.derivedFunctions[name]) then
			local func = t.derivedFunctions[name]
			return function(...) return func(t,...) end
		end

		return self[name]
	end})

	if( not rawget(stats, "derivedStats")) then
		stats.derivedStats = self.derivedStats
	end

	if( not rawget(stats, "derivedFunctions")) then
		stats.derivedFunctions = self.derivedFunctions
	end

	stats.relevant = {}
	return stats
end


function CatPoints:CreateTable(displayName,  stats)

	local s = Player:new(stats, displayName)

	s.relevant = {}

	self.Modifiers[displayName]=s
	return stats
end


function CatPoints:FillTalents()
	for t = 1, GetNumTalentTabs() do
		for i = 1, GetNumTalents(t) do
			local name, _, _, _, currRank = GetTalentInfo(t, i)
			if(currRank > 0) then
				name = name:gsub("%W", "")
				Player[name]=currRank
			end
		end
	end
end

function CatPoints:CombineBonuses(new, old)
	local total = {}
	for k,v in pairs(new) do
		total[k]=v
	end

	for k,v in pairs(old) do
		total[k] = (total[k] or 0) - v
	end

	return total
end



function CatPoints:ArmorFromMitigation(mitigation, penetration)
	if(mitigation < 0 or mitigation > .75) then
		error(format("Mitigation (%.2f) is not in the range [0...75]", mitigation))
	end
	local armor= (467.5 * 73 - 22167.5) / (1 / mitigation- 1) + (penetration or 0)
	if(armor < 0) then
		armor = 0
	end

	return armor
end

function CatPoints:MitigationFromArmor(armor)
	local mitigation = (armor / (armor + 467.5 * 73 - 22167.5 ))
	if(mitigation > .75) then
		mitigation = .75
	end
	return mitigation
end

function CatPoints:LinkSet(link)
	local bonus = self.Bonus:ScanItemLink(link, true)
	if(bonus and bonus.set) then
		local b = bonus.set:gsub("%W", "")
		bonus.bonuses[b]=1
		return b,bonus.bonuses
	end
	return nil, bonus.bonuses
end


function CatPoints:UpdateSets()
	local set= {}

	for x=1,18 do
		local equippedLink = GetInventoryItemLink("player",x)
		if(equippedLink) then
			local setName = self:LinkSet(equippedLink)
			if(setName) then
				set[setName] = (set[setName] or 0) + 1
			end
		end
	end
	return set
end


CatPoints.Modifiers = {}


function CatPoints:AddCatPoints(frame, link)

	if(frame and link) then
		local _, _, _, _, _, _, _, _,itemEquipLoc  = GetItemInfo(link)
		if(itemEquipLoc and invType[itemEquipLoc]) then
			local idol = (itemEquipLoc == 'INVTYPE_RELIC' and self:GetIdol(link)) or nil

			itemEquipLoc = invType[itemEquipLoc]
			local equippedLink = BareID(GetInventoryItemLink("player",itemEquipLoc))
			if(itemCache.eqlink ~= equippedLink or itemCache.link ~= link) then
				itemCache.eqlink = equippedLink
				itemCache.link = link
				itemCache.lines = {}

				local eqBonuses = {}
				if(equippedLink and equippedLink ~= link) then
					_,eqBonuses =  self:LinkSet(equippedLink)
				end


				local _, itemBonuses = self:LinkSet(link)
				self:MergeSpecial(itemBonuses, link)
				self:MergeSpecial(eqBonuses, equippedLink)
				local netBonus = self:CombineBonuses(itemBonuses or {},eqBonuses)


				if(idol) then
					netBonus.Idol = idol
				end

				for k,v in pairs(self.Modifiers) do
					if(v and v:display()) then
						local b, sb = v:compare(netBonus)
						self:AddPointsLine(frame, k, b, sb, true)
					end
				end
			else
				for _,v in ipairs(itemCache.lines) do
					self:AddPointsLine(frame, v[1], v[2])
				end
			end
		else
			-- must be an enchant or gem

			if( itemCache.link ~= link) then
				itemCache.link = link
				itemCache.lines = {}
				local bonuses  = CatPoints.Bonus:ScanItem(link, true) or {}

				local equippedBonuses = {}

				--bonuses.Idol = self:GetIdol(link)

				for k,v in pairs(self.Modifiers) do
					if(v and v:display()) then
						local b, sb = v:compare(bonuses)
						self:AddPointsLine(frame, k, b, sb, true)
					end
				end
			else
				for _,v in ipairs(itemCache.lines) do
					self:AddPointsLine(frame, v[1], v[2])
				end
			end
		end
		frame:Show()
	end
end

function CatPoints:UNIT_INVENTORY_CHANGED(units)
	if units=='player' then
		self.EquippedSets = self:UpdateSets()
	end
end


function CatPoints:OnInitialize()

end

function CatPoints:OnEnable(first)
	if(first) then

		self:RegisterEvent("UNIT_INVENTORY_CHANGED")

		self.EquippedSets = self:UpdateSets()

		self:FillTalents()

		self:SecureHook(GameTooltip, "SetBagItem", function(this, bag, slot)
				self:DoTooltip(GameTooltip, GetContainerItemLink(bag, slot));
			end
		)

		self:SecureHook(GameTooltip, "SetLootItem", function(this, slot)
				self:DoTooltip(GameTooltip, GetLootSlotLink(slot));
			end
		)

		self:SecureHook(GameTooltip, "SetQuestItem", function(this, unit, slot)
				self:DoTooltip(GameTooltip, GetQuestItemLink(unit, slot));
			end
		)

		self:SecureHook(GameTooltip, "SetQuestLogItem", function(this, sOpt, slot)
				self:DoTooltip(GameTooltip, GetQuestLogItemLink(sOpt, slot));
			end
		)
		self:SecureHook(GameTooltip, "SetTradeSkillItem", function(this, skill, slot)
				local link = (slot) and GetTradeSkillReagentItemLink(skill, slot) or GetTradeSkillItemLink(skill);
				self:DoTooltip(GameTooltip, link);
			end
		)
		self:SecureHook(GameTooltip, "SetMerchantItem", function(this, slot)
				self:DoTooltip(GameTooltip, GetMerchantItemLink(slot));
			end
		)
		self:SecureHook(GameTooltip, "SetAuctionItem", function(this, unit, slot)
				self:DoTooltip(GameTooltip, GetAuctionItemLink(unit, slot));
			end
		)

		self:SecureHook(GameTooltip, "SetLootRollItem", function(this, id)
				self:DoTooltip(GameTooltip, GetLootRollItemLink(id));
			end
		)
		self:SecureHook(GameTooltip, "SetHyperlink", function(this, link)
				self:DoTooltip(GameTooltip, link);
			end
		)
		self:Hook(GameTooltip, "SetInventoryItem", function(this, unit, slot)
				local sItem, sCooldown, sRepair = self.hooks[GameTooltip]["SetInventoryItem"](this, unit, slot);
				if(not sItem) then return nil end
				--if (slot > 39) then -- bank
					self:DoTooltip(GameTooltip, GetInventoryItemLink(unit, slot));
				--end
				return sItem, sCooldown, sRepair;
			end
		, true)
		self:SecureHook("SetItemRef", function(link, name, button)
				if(link and name and ItemRefTooltip) then
					if( strsub(link, 1, 6) ~= "Player" ) then
						if( ItemRefTooltip:IsVisible()) then
							if(not DressUpFrame:IsVisible()) then
								self:DoTooltip(ItemRefTooltip, link);
							end
							ItemRefTooltip.isDisplayDone = nil;
						end
					end
				end
			end
  )
	end
end
function CatPoints:DoTooltip(frame, link)
	link = BareID(link)
	self:AddCatPoints(frame, link)
end


function CatPoints:Equivalence(name, noDisplay)
	if(self.Modifiers[name]) then
		if( self.Modifiers[name].equivalence) then
			self.Modifiers[name]:equivalence(noDisplay)
		else
			self:Print(name.." doesn't define a equivalence function")
		end
	else
		self:Print("Unable to find a stat named "..name)
	end
end




local options = {
	type = 'group',
	args = {
		equiv = {
			name = 'equiv',
			type = 'text',
			desc = "Print equivalence values",
			usage = "<name of point values to compute>",
			set = 'Equivalence',
			get = false
		}
	}
}

CatPoints:RegisterChatCommand("/catpoints", "/cp", options)



